import copy

import torch
import torch.nn.functional as F
from torch.nn import CrossEntropyLoss, MSELoss
from torch_geometric.data import Batch, Data
from torch_geometric.explain.explanation import Explanation
from torch_geometric.graphgym.config import cfg
from torch_geometric.utils import add_random_edge, dropout_edge, dropout_node

from explaining_framework.metric.base import Metric
from explaining_framework.utils.io import obj_config_to_str


def compute_gradient(model, inp, target, loss):
    with torch.autograd.set_grad_enabled(True):
        inp.x.requires_grad = True
        out = model(x=inp.x, edge_index=inp.edge_index)
        err = loss(out, target)
        return torch.autograd.grad(err, inp.x)[0]


class Attack(Metric):
    def __init__(
        self,
        name: str,
        model: torch.nn.Module,
        dropout: float = 0.5,
        loss: torch.nn = None,
    ):

        super().__init__(name=name, model=model)
        self.name = name
        self.model = model
        self.authorized_metric = [
            "gaussian_noise",
            "add_edge",
            "remove_edge",
            "remove_node",
            "pgd",
            "fgsm",
            "no_attack",
        ]
        self.dropout = dropout
        self.config = None
        if loss is None:
            if cfg.model.loss_fun == "cross_entropy":
                self.loss = CrossEntropyLoss()
            elif cfg.model.loss_fun == "mse":
                self.loss = MSELoss()
            else:
                raise ValueError(f"{loss} is not supported yet")
        else:
            self.loss = loss
        self.load_metric(name)

    def _gaussian_noise(self, exp) -> Explanation:
        exp_ = exp.clone()
        exp_.x = exp_.x + torch.randn(*exp_.x.shape).to(exp_.x.device)
        return exp_

    def _add_edge(self, exp, p: float) -> Explanation:
        exp_ = exp.clone()
        exp_.edge_index, _ = add_random_edge(
            exp_.edge_index, p=p, num_nodes=exp_.x.shape[0]
        )
        return exp_

    def _remove_edge(self, exp, p: float) -> Explanation:
        exp_ = exp.clone()
        exp_.edge_index, _ = dropout_edge(exp_.edge_index, p=p)
        return exp_

    def _remove_node(self, exp, p: float) -> Explanation:
        exp_ = exp.clone()
        exp_.edge_index, _, _ = dropout_node(
            exp_.edge_index, p=p, num_nodes=exp_.x.shape[0]
        )
        return exp_

    def _load_add_edge(self):
        return lambda exp: self._add_edge(exp, p=self.dropout)

    def _load_remove_edge(self):
        return lambda exp: self._remove_edge(exp, p=self.dropout)

    def _load_remove_node(self):
        return lambda exp: self._remove_node(exp, p=self.dropout)

    def _load_gaussian_noise(self):
        return lambda exp: self._gaussian_noise(exp)

    def _load_no_attack(self):
        return lambda exp: exp

    def load_metric(self, name):
        if name in self.authorized_metric:
            if name == "gaussian_noise":
                self.metric = self._load_gaussian_noise()
            if name == "no_attack":
                self.metric = self._load_no_attack()
            if name == "add_edge":
                self.metric = self._load_add_edge()
            if name == "remove_edge":
                self.metric = self._load_remove_edge()
            if name == "remove_node":
                self.metric = self._load_remove_node()
            if name == "pgd":
                pgd = PGD(
                    model=self.model,
                    loss=self.loss,
                    epsilon=1,
                    radius=1,
                    step_num=50,
                    random_start=False,
                    norm="inf",
                )
                self.config = obj_config_to_str(pgd.__dict__)
                self.metric = lambda exp: pgd.forward(
                    input=exp,
                    target=exp.y,
                )
            if name == "fgsm":
                fgsm = FGSM(model=self.model, loss=self.loss, epsilon=1)
                self.config = obj_config_to_str(fgsm.__dict__)
                self.metric = lambda exp: fgsm.forward(input=exp, target=exp.y)
        else:
            raise ValueError(f"{name} is not supported yet")

        return self.metric

    def forward(self, exp) -> Explanation:
        attack = self.metric(exp)
        return attack

    def get_attacked_prediction(self, data: Data) -> Data:
        data_ = data.clone()
        data_attacked = self.forward(data_)
        pred = self.get_prediction(x=data_.x, edge_index=data_.edge_index)
        pred_attacked = self.get_prediction(
            x=data_attacked.x, edge_index=data_attacked.edge_index
        )
        setattr(data_, "pred", pred)
        setattr(data_, "pred_attacked", pred_attacked)
        return data_


class FGSM(Metric):
    def __init__(
        self,
        model: torch.nn.Module,
        loss: torch.nn.Module,
        lower_bound: float = float("-inf"),
        upper_bound: float = float("inf"),
        epsilon=1,
    ):
        super().__init__(name="fgsm", model=model)
        self.model = model
        self.loss = loss
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.epsilon = epsilon

        self.bound = lambda x: torch.clamp(
            x,
            min=torch.Tensor([lower_bound]).to(x.device),
            max=torch.Tensor([upper_bound]).to(x.device),
        ).to(x.device)

        self.zero_thresh = 10**-6

    def forward(self, input, target) -> Explanation:
        input_ = input.clone()
        grad = compute_gradient(
            model=self.model, inp=input_, target=target, loss=self.loss
        )
        grad = self.bound(grad)
        input_.x = torch.where(
            torch.abs(grad) > self.zero_thresh,
            input_.x - self.epsilon * torch.sign(grad),
            input_.x,
        )
        return input_

    def load_metric(self):
        pass


class PGD(Metric):
    def __init__(
        self,
        model: torch.nn.Module,
        loss: torch.nn.Module,
        lower_bound: float = float("-inf"),
        upper_bound: float = float("inf"),
        epsilon=1,
        radius=1,
        step_num=50,
        random_start=False,
        norm="inf",
    ):
        super().__init__(name="pgd", model=model)
        self.model = model
        self.loss = loss
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.bound = lambda x: torch.clamp(
            x,
            min=torch.Tensor([lower_bound]).to(x.device),
            max=torch.Tensor([upper_bound]).to(x.device),
        ).to(x.device)

        self.zero_thresh = 10**-6
        self.fgsm = FGSM(
            model=model, loss=loss, lower_bound=lower_bound, upper_bound=upper_bound
        )
        self.epsilon = epsilon
        self.radius = radius
        self.step_num = step_num
        self.random_start = random_start
        self.norm = norm

    def forward(
        self,
        input,
        target,
    ) -> Explanation:
        def _clip(inputs: Explanation, outputs: Explanation) -> Explanation:
            diff = outputs.x - inputs.x
            if self.norm == "inf":
                inputs.x = inputs.x + torch.clamp(diff, -self.radius, self.radius)
                return inputs
            elif self.norm == "2":
                inputs.x = inputs.x + torch.renorm(diff, 2, 0, self.radius)
                return inputs
            else:
                raise AssertionError("Norm constraint must be L2 or Linf.")

        perturbed_input = input
        if self.random_start:
            perturbed_input = self.bound(
                self._random_point(input.x, self.radius, self.norm)
            )
        for _ in range(self.step_num):
            perturbed_input = self.fgsm.forward(input=perturbed_input, target=target)
            perturbed_input = _clip(input, perturbed_input)
            perturbed_input.x = self.bound(perturbed_input.x).detach()
        return perturbed_input

    def load_metric(self):
        pass
